import unittest
import json
from src.parrotflume.tools import handle_tool_call


class TestHandleToolCall(unittest.TestCase):
    def setUp(self):
        # Initialize a messages list for each test
        self.messages = []

    def test_handle_get_current_date(self):
        # Test handling a get_current_date tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'get_current_date',
                'arguments': json.dumps({})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "get_current_date")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

        # Verify the content is a valid date in the format YYYY-MM-DD
        from datetime import datetime
        try:
            datetime.strptime(self.messages[0]["content"], "%Y-%m-%d")
            date_is_valid = True
        except ValueError:
            date_is_valid = False
        self.assertTrue(date_is_valid, "The date format is invalid")

    def test_handle_sympy_simplify(self):
        # Test handling a sympy_simplify tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'sympy_simplify',
                'arguments': json.dumps({"expression": "x + x"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "sympy_simplify")
        self.assertEqual(self.messages[0]["content"], "2*x")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_sympy_solve(self):
        # Test handling a sympy_solve tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'sympy_solve',
                'arguments': json.dumps({"expression": "x**2 - 4", "variable": "x"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "sympy_solve")
        self.assertEqual(self.messages[0]["content"], "[-2, 2]")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_sympy_integrate(self):
        # Test handling a sympy_integrate tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'sympy_integrate',
                'arguments': json.dumps({"expression": "x**2", "variable": "x"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "sympy_integrate")
        self.assertEqual(self.messages[0]["content"], "x**3/3")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_sympy_differentiate(self):
        # Test handling a sympy_differentiate tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'sympy_differentiate',
                'arguments': json.dumps({"expression": "x**2", "variable": "x"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "sympy_differentiate")
        self.assertEqual(self.messages[0]["content"], "2*x")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_regex_match(self):
        # Test handling a regex_match tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'regex_match',
                'arguments': json.dumps({"pattern": r"\d+", "text": "123 abc 456"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "regex_match")
        self.assertEqual(self.messages[0]["content"], "['123', '456']")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_count_chars(self):
        # Test handling a count_chars tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'count_chars',
                'arguments': json.dumps({"text": "1234567890"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "count_chars")
        self.assertEqual(self.messages[0]["content"], "10")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")

    def test_handle_count_substring_occurrences(self):
        # Test handling a count_substring_occurrences tool call
        tool_call = type('ToolCall', (), {
            'id': 'tool_call_123',
            'function': type('Function', (), {
                'name': 'count_substring_occurrences',
                'arguments': json.dumps({"string": "strawberry", "substring": "r"})
            })
        })

        # Call the function
        handle_tool_call(self.messages, tool_call)

        # Verify the result
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0]["role"], "tool")
        self.assertEqual(self.messages[0]["name"], "count_substring_occurrences")
        self.assertEqual(self.messages[0]["content"], "3")
        self.assertEqual(self.messages[0]["tool_call_id"], "tool_call_123")


if __name__ == "__main__":
    unittest.main()


import json
import re
from datetime import datetime
from sympy import simplify, solve, sympify, Eq, integrate, diff
from sympy.parsing.sympy_parser import parse_expr

tools = [
    {
        "type": "function",
        "function": {
            "name": "get_current_date",
            "description": "Get the current (today's) date",
            "parameters": {
                "type": "object",
                "properties": {},
                "required": []
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "sympy_simplify",
            "description": "Simplify a mathematical expression using sympy.simplify",
            "parameters": {
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "mathematical expression to simplify, using sympy syntax"
                    }
                },
                "required": ["expression"]
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "sympy_solve",
            "description": "Solve a mathematical expression using sympy.solve",
            "parameters": {
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "mathematical expression to solve, using sympy syntax"
                    },
                    "variable": {
                        "type": "string",
                        "description": "variable to solve for, using sympy syntax"
                    }
                },
                "required": ["expression", "variable"]
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "sympy_integrate",
            "description": "Integrate a mathematical expression using sympy.integrate",
            "parameters": {
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "mathematical expression to integrate, using sympy syntax"
                    },
                    "variable": {
                        "type": "string",
                        "description": "variable of integration, using sympy syntax"
                    }
                },
                "required": ["expression", "variable"]
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "sympy_differentiate",
            "description": "Differentiate a mathematical expression using sympy.diff",
            "parameters": {
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "mathematical expression to differentiate, using sympy syntax"
                    },
                    "variable": {
                        "type": "string",
                        "description": "variable of differentiation, using sympy syntax"
                    },
                    "order": {
                        "type": "integer",
                        "description": "order of differentiation (optional, default is 1)"
                    }
                },
                "required": ["expression", "variable"]
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "regex_match",
            "description": "Applies a regex pattern to a text and returns the matches",
            "parameters": {
                "type": "object",
                "properties": {
                    "pattern": {
                        "type": "string",
                        "description": "regex pattern to apply."
                    },
                    "text": {
                        "type": "string",
                        "description": "text to search within."
                    }
                },
                "required": ["pattern", "text"]
            }
        }
    },
    {
        "type": "function",
        "function": {
            "name": "count_chars",
            "description": "Counts the number of characters in a string",
            "parameters": {
                "type": "object",
                "properties": {
                    "text": {
                        "type": "string",
                        "description": "string whose characters are to be counted",
                    }
                },
                "required": ["text"],
            },
        },
    },
    {
        "type": "function",
        "function": {
            "name": "count_substring_occurrences",
            "description": "Counts the number of occurrences of a substring in a string",
            "parameters": {
                "type": "object",
                "properties": {
                    "string": {
                        "type": "string",
                        "description": "string to search within"
                    },
                    "substring": {
                        "type": "string",
                        "description": "substring to count occurrences of"
                    }
                },
                "required": ["string", "substring"]
            }
        }
    },
]


def handle_get_current_date(messages):
    result = datetime.now().strftime("%Y-%m-%d")
    messages.append({"role": "tool", "name": "get_current_date", "content": result})


def handle_sympy_simplify(messages, arguments):
    try:
        expr = parse_expr(arguments["expression"])
        simplified_expr = simplify(expr)
        messages.append({"role": "tool", "name": "sympy_simplify", "content": str(simplified_expr)})
    except Exception as e:
        messages.append({"role": "tool", "name": "sympy_simplify", "content": str(e)})


def handle_sympy_solve(messages, arguments):
    try:
        equation = arguments["expression"]
        variable = arguments["variable"]

        if "=" in equation:
            left, right = equation.split("=", 1)
            left_expr = parse_expr(left.strip())
            right_expr = parse_expr(right.strip())
            eq = Eq(left_expr, right_expr)
        else:
            eq = parse_expr(equation)

        solution = solve(eq, sympify(variable))
        messages.append({"role": "tool", "name": "sympy_solve", "content": str(solution)})
    except Exception as e:
        messages.append({"role": "tool", "name": "sympy_solve", "content": str(e)})


def handle_sympy_integrate(messages, arguments):
    try:
        expression = arguments["expression"]
        variable = sympify(arguments["variable"])

        expr = parse_expr(expression)
        result = integrate(expr, variable)
        messages.append({"role": "tool", "name": "sympy_integrate", "content": str(result)})
    except Exception as e:
        messages.append({"role": "tool", "name": "sympy_integrate", "content": str(e)})


def handle_sympy_differentiate(messages, arguments):
    try:
        expression = arguments["expression"]
        variable = sympify(arguments["variable"])
        order = arguments.get("order", 1)

        expr = parse_expr(expression)
        result = diff(expr, variable, order)
        messages.append({"role": "tool", "name": "sympy_differentiate", "content": str(result)})
    except Exception as e:
        messages.append({"role": "tool", "name": "sympy_differentiate", "content": str(e)})


def handle_regex_match(messages, arguments):
    try:
        pattern = arguments["pattern"]
        text = arguments["text"]

        result = re.findall(pattern, text)
        messages.append({"role": "tool", "name": "regex_match", "content": str(result)})
    except Exception as e:
        messages.append({"role": "tool", "name": "regex_match", "content": str(e)})


def handle_count_chars(messages, arguments):
    try:
        text = arguments["text"]

        result = len(text)
        messages.append({"role": "tool", "name": "count_chars", "content": str(result)})
    except Exception as e:
        messages.append({"role": "tool", "name": "count_chars", "content": str(e)})


def handle_count_substring_occurrences(messages, arguments):
    try:
        text = arguments["string"]
        substring = arguments["substring"]

        result = text.count(substring)
        messages.append({"role": "tool", "name": "count_substring_occurrences", "content": str(result)})
    except Exception as e:
        messages.append({"role": "tool", "name": "count_substring_occurrences", "content": str(e)})


def handle_tool_call(messages, tool_call, do_print=False):
    if tool_call.function.arguments:
        args = json.loads(tool_call.function.arguments)
    else:
        # Claude via openrouter sends an empty string as argument for argument-less tool call
        args = None
    if do_print:
        print(f"[{tool_call.function.name} called]")

    if tool_call.function.name == "get_current_date":
        result = datetime.now().strftime("%Y-%m-%d")
        messages.append({
            "role": "tool",
            "tool_call_id": tool_call.id,
            "name": "get_current_date",
            "content": result
        })
    elif tool_call.function.name == "sympy_simplify":
        try:
            expr = parse_expr(args["expression"])
            simplified_expr = simplify(expr)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_simplify",
                "content": str(simplified_expr)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_simplify",
                "content": str(e)
            })
    elif tool_call.function.name == "sympy_solve":
        try:
            equation = args["expression"]
            variable = args["variable"]

            if "=" in equation:
                left, right = equation.split("=", 1)
                left_expr = parse_expr(left.strip())
                right_expr = parse_expr(right.strip())
                eq = Eq(left_expr, right_expr)
            else:
                eq = parse_expr(equation)

            solution = solve(eq, sympify(variable))
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_solve",
                "content": str(solution)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_solve",
                "content": str(e)
            })
    elif tool_call.function.name == "sympy_integrate":
        try:
            expression = args["expression"]
            variable = sympify(args["variable"])

            expr = parse_expr(expression)
            result = integrate(expr, variable)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_integrate",
                "content": str(result)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_integrate",
                "content": str(e)
            })
    elif tool_call.function.name == "sympy_differentiate":
        try:
            expression = args["expression"]
            variable = sympify(args["variable"])
            order = args.get("order", 1)

            expr = parse_expr(expression)
            result = diff(expr, variable, order)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_differentiate",
                "content": str(result)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "sympy_differentiate",
                "content": str(e)
            })
    elif tool_call.function.name == "regex_match":
        try:
            pattern = args["pattern"]
            text = args["text"]

            result = re.findall(pattern, text)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "regex_match",
                "content": str(result)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "regex_match",
                "content": str(e)
            })
    elif tool_call.function.name == "count_chars":
        try:
            text = args["text"]

            result = len(text)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "count_chars",
                "content": str(result)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "count_chars",
                "content": str(e)
            })
    elif tool_call.function.name == "count_substring_occurrences":
        try:
            text = args["string"]
            substring = args["substring"]

            result = text.count(substring)
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "count_substring_occurrences",
                "content": str(result)
            })
        except Exception as e:
            messages.append({
                "role": "tool",
                "tool_call_id": tool_call.id,
                "name": "count_substring_occurrences",
                "content": str(e)
            })



import re

from pylatexenc import latexwalker, latex2text
from pylatexenc.macrospec import MacroSpec

#  Overwrite sqrt macros from pylatexenc, the original wrongly renders n-th roots as square roots
lw_context_db = latexwalker.get_default_latex_context_db()
lw_context_db.add_context_category(
    'roots',
    prepend=True,
    macros=[
        MacroSpec("sqrt", "[{")
    ],
)

l2t_context_db = latex2text.get_default_latex_context_db()
l2t_context_db.add_context_category(
    'roots',
    prepend=True,
    macros=[
        latex2text.MacroTextSpec("sqrt", simplify_repl="%(1)s√(%(2)s)"),
    ],
)

# ANSI escape codes
BOLD = "\033[1m"
ITALIC = "\033[3m"
INVERT = "\033[7m"  # Inversion for h3 headers
UNDERLINE = "\033[4m"  # Underline for h4 headers
GREY_BG = "\033[100m"  # Grey background for code blocks
FAINT = "\033[2m"  # Dim for code block headers

RESET_BOLD = "\033[22m"
RESET_ITALIC = "\033[23m"
RESET_INVERT = "\033[27m"
RESET_UNDERLINE = "\033[24m"
RESET_GREY_BG = "\033[49m"
RESET_FAINT = "\033[22m"

RESET_FG = "\033[39m"
RESET_GENERIC = "\033[0m"

fg_color_map = {
    "black": "\033[30m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[37m",
    "bright_black": "\033[1;30m",
    "bright_red": "\033[1;31m",
    "bright_green": "\033[1;32m",
    "bright_yellow": "\033[1;33m",
    "bright_blue": "\033[1;34m",
    "bright_magenta": "\033[1;35m",
    "bright_cyan": "\033[1;36m",
    "bright_white": "\033[1;37m",
}


def print_fancy(text, do_markdown, do_latex, do_color, color):
    """
    Processes the given text with Markdown, LaTeX, and color formatting.
    Supports h3, h4, italic (*), bold (**), code blocks, inline code, and LaTeX to Unicode conversion.
    """
    if do_latex:
        do_latex = re.search(r'\\[a-zA-Z]', text)  # Probably no latex in here, otherwise
    result = []
    in_code_block = False
    in_latex_block = False
    latex_block = []

    for line in text.splitlines():
        # Handle code blocks
        if line.lstrip().startswith("\x60\x60\x60"):
            line = line.lstrip()
            if not in_code_block:
                in_code_block = True
                result.append(FAINT + line[3:] + RESET_FAINT + GREY_BG + "\n")
            else:
                in_code_block = False
                result.append(line[3:] + RESET_GREY_BG + "\n")
            continue

        if not in_code_block:
            # Handle LaTeX matrices
            if do_latex:
                if line.strip() == "\\[":
                    in_latex_block = True
                    latex_block = []
                    continue
                elif line.strip() == "\\]":
                    in_latex_block = False
                    latex_text = "\n".join(latex_block)
                    latex_text = custom_latex_to_text(latex_text)
                    latex_text = latex_text.replace(";", "\n ")
                    result.append(latex_text + "\n")
                    continue

                if in_latex_block:
                    latex_block.append(line)
                    continue

            if do_markdown:
                # Render header h3
                if line.startswith("### "):
                    line = INVERT + line[4:] + RESET_INVERT
                # Render header h4
                elif line.startswith("#### "):
                    line = UNDERLINE + line[5:] + RESET_UNDERLINE

                # Render inline bold (**)
                try:
                    line = re.sub(r"\*\*(.*?)\*\*", lambda m: BOLD + (m.group(1) or m.group(2)) + RESET_BOLD, line)
                except IndexError:
                    pass  # ignore

                # Render inline italic (*)
                try:
                    line = re.sub(r"\*(.*?)\*", lambda m: ITALIC + (m.group(1) or m.group(2)) + RESET_ITALIC, line)
                except IndexError:
                    pass  # ignore

                # Render inline code (`)
                try:
                    line = re.sub(r"`(.*?)`", lambda m: GREY_BG + (m.group(1) or m.group(2)) + RESET_GREY_BG, line)
                except IndexError:
                    pass  # ignore

            if do_latex and not re.search(r"`.*?`", line):
                # Convert LaTeX to Unicode
                line = custom_latex_to_text(line)

            if do_color:
                # Apply yellow foreground color
                line = fg_color_map.get(color, RESET_FG) + line + RESET_FG

        result.append(line + "\n")

    print("".join(result), end=None)


def print_reset():
    print(RESET_GENERIC, end=None)


def custom_latex_to_text(input_latex):
    # the latex parser instance with custom latex_context
    lw_obj = latexwalker.LatexWalker(input_latex, latex_context=lw_context_db, tolerant_parsing=True)
    # parse to node list
    nodelist, pos, length = lw_obj.get_latex_nodes()
    # initialize the converter to text with custom latex_context
    l2t_obj = latex2text.LatexNodes2Text(latex_context=l2t_context_db, keep_comments=True)
    # convert to text
    try:
        return l2t_obj.nodelist_to_text(nodelist)
    except ValueError:
        return input_latex



import os
import readline
from functools import lru_cache
import openai


class AutoCompleter:
    def __init__(self, api_providers):
        self.api_providers = api_providers

    def auto_complete(self, text, state):
        line = readline.get_line_buffer()

        # Handle API provider completion for "/p " commands
        if line.startswith("/p "):
            # Extract the prefix after "/p "
            prefix = line[3:].lstrip()

            # Filter providers that match the prefix
            matches = [provider for provider in self.api_providers if provider.startswith(prefix)]

            # Return the match corresponding to the state
            if state < len(matches):
                return matches[state]
            else:
                return None

        # Handle model completion for "/m " commands
        elif line.startswith("/m "):
            # Extract the prefix after "/m "
            prefix = line[3:].lstrip()

            # Get available models from the OpenAI API
            model_ids = self.get_models(openai.base_url)

            # Filter models that match the prefix
            matches = [model for model in model_ids if model.startswith(prefix)]

            # Return the match corresponding to the state
            if state < len(matches):
                return matches[state]
            else:
                return None

        # Check if the input starts with a file command
        if not any(line.startswith(f"/{cmd} ") for cmd in ("c", "d", "f", "u")):
            return None

        # Expand ~ to the user's home directory
        if '~' in text:
            text = os.path.expanduser(text)

        # Get the directory and prefix
        directory, prefix = os.path.split(text)

        # If no directory is specified, use the current directory
        if not directory:
            directory = '.'

        # Get all files and directories in the specified directory
        try:
            files = os.listdir(directory)
        except OSError:
            return None

        # Filter files that match the prefix
        matches = [f for f in files if f.startswith(prefix)]

        # Add the directory back to the matches
        matches = [os.path.join(directory, f) for f in matches]

        # Return the match corresponding to the state
        if state < len(matches):
            return matches[state]
        else:
            return None

    def setup(self):
        readline.set_completer(self.auto_complete)
        readline.parse_and_bind("tab: complete")
        readline.set_completer_delims(' \t\n')  # Treat spaces and tabs as delimiters

    @lru_cache
    def get_models(self, _cache_key):
        # noinspection PyBroadException
        try:
            return [model.id for model in openai.models.list()]
        except Exception:
            return []



no_function_call = ["o1-preview"]  # supports no function calling at all
no_function_call_looping = ["deepseek-"]  # runs into endless loops if looping is allowed
no_system = ["o1-preview"]  # does not know role "system"
developer = ["o1", "o3"]  # "system" has been renamed to "developer"
max_completion_tokens = ["o1", "o3"]  # uses "max_completion_tokens" instead of "max_tokens"
no_temperature = ["o1", "o3"]  # has no temperature setting



import unittest
from io import StringIO
import contextlib
from src.parrotflume.fancy import print_fancy, print_reset

class TestPrintFancy(unittest.TestCase):
    @staticmethod
    def capture_output(func, *args, **kwargs):
        """Helper function to capture stdout from a function."""
        captured_output = StringIO()
        with contextlib.redirect_stdout(captured_output):
            func(*args, **kwargs)
        return captured_output.getvalue().strip()

    def test_markdown_formatting(self):
        # Test combined Markdown formatting (headers, bold, italic, code blocks)
        text = (
            "### Header H3\n"
            "#### Header H4\n"
            "**Bold text**\n"
            "*Italic text*\n"
            "`inline code`\n"
            "```bash\n"
            "echo \"code block\"\n"
            "```"
        )
        output = self.capture_output(print_fancy, text, do_markdown=True, do_latex=False, do_color=False, color="bright_yellow")

        # Expected output with ANSI escape codes for Markdown formatting
        expected_output = (
            "\033[7mHeader H3\033[27m\n"
            "\033[4mHeader H4\033[24m\n"
            "\033[1mBold text\033[22m\n"
            "\033[3mItalic text\033[23m\n"
            "\033[100minline code\033[49m\n"
            "\033[2mbash\033[22m\033[100m\n"
            "echo \"code block\"\n"
            "\033[49m"
        )
        self.assertEqual(expected_output, output)

    def test_latex_conversion(self):
        # Test complex LaTeX-to-Unicode conversion
        text = r"\sqrt{x} + \frac{a}{b} = \int_{0}^{1} f(x) \, dx"
        output = self.capture_output(print_fancy, text, do_markdown=False, do_latex=True, do_color=False, color="bright_yellow")

        # Expected output after LaTeX-to-Unicode conversion
        expected_output = "√(x) + a/b = ∫_0^1 f(x)   dx"
        self.assertEqual(expected_output, output)

    def test_color_formatting(self):
        # Test color formatting
        output = self.capture_output(print_fancy, "Hello, World!", do_markdown=False, do_latex=False, do_color=True, color="bright_yellow")

        # Expected output with ANSI escape codes for bright yellow
        expected_output = "\033[1;33mHello, World!\033[39m"
        self.assertEqual(expected_output, output)

    def test_print_reset(self):
        # Test print_reset function
        output = self.capture_output(print_reset)

        # Expected output: ANSI reset code
        expected_output = "\033[0m"
        self.assertEqual(expected_output, output)


if __name__ == "__main__":
    unittest.main()


from dataclasses import dataclass


@dataclass
class Config:
    api_provider: str = None
    base_url: str = None
    api_key: str = None
    model: str = None
    temperature: float = 0.0
    max_tokens: int = 4096
    do_markdown: bool = True
    do_latex: bool = True
    do_color: bool = False
    color: str = "bright_yellow"
    func: bool = True
    json: bool = False
    prefix: str = None


base_urls = {
    "openai" : "https://api.openai.com/v1/",
    "deepseek" : "https://api.deepseek.com/v1/",
    "llama.cpp" : "http://localhost:8080/v1/",
    "openrouter" : "https://openrouter.ai/api/v1/"
}
api_keys = {
    "llama.cpp" : "sk-no-key-required"
}
models = {
    "openai" : "gpt-4o",
    "deepseek" : "deepseek-chat",
    "llama.cpp" : ""
}



import argparse
import base64
import json
import os
# noinspection PyUnresolvedReferences
import readline  # used by input()
import sys
import time
import tomllib
import openai
from openai.types.chat import ChatCompletionMessageToolCall
from appdirs import user_config_dir

from parrotflume.tools import tools, handle_tool_call
from parrotflume.fancy import print_fancy, print_reset
from parrotflume.auto_completer import AutoCompleter
from parrotflume.config import Config
from parrotflume import fallbacks
from parrotflume import model_quirks
from parrotflume import __version__

app_name = "parrotflume"
config_dir = user_config_dir(appname=app_name)
config_path = os.path.join(config_dir, f"{app_name}.config.toml")


def load_config(file_path):
    try:
        with open(file_path, "rb") as f:
            config_data = tomllib.load(f)
            return config_data
    except FileNotFoundError:
        try:
            os.makedirs(config_dir, exist_ok=True)
        except OSError:
            pass  # ignore
        return None
    except tomllib.TOMLDecodeError as e:
        print(f"Error parsing TOML config file: {e}", file=sys.stderr)
        sys.exit(1)


def truncate_tool_call_ids(messages, max_id_length=40):
    for message in messages:
        # the call
        if "tool_calls" in message:
            for tool_call in message["tool_calls"]:
                # tool_call is an object
                if hasattr(tool_call, "id") and len(tool_call.id) > max_id_length:
                    tool_call.id = tool_call.id[:max_id_length]
                # tool_call is serialized
                elif "id" in tool_call and len(tool_call["id"]) > max_id_length:
                    tool_call["id"] = tool_call["id"][:max_id_length]
        # the return
        elif "tool_call_id" in message and len(message["tool_call_id"]) > max_id_length:
            message["tool_call_id"] = message["tool_call_id"][:max_id_length]
    return messages


def create_completion_response(config, messages, add_functions):
    # Ugly quirk for openAI o1-preview* model: Does not know system messages
    if any(config.model.startswith(prefix) for prefix in model_quirks.no_system):
        if messages[0]["role"] == "system" or messages[0]["role"] == "developer":
            messages[0]["role"] = "user"
    # Ugly quirk for openAI o* (non-preview) models: system has been renamed to developer
    elif any(config.model.startswith(prefix) for prefix in model_quirks.developer):
        if messages[0]["role"] == "system":
            messages[0]["role"] = "developer"
    else:
        if messages[0]["role"] in "developer":
            messages[0]["role"] = "system"

    # Ugly quirk for openAI o1-preview* model: Does not know function calls
    if not config.func or any(config.model.startswith(prefix) for prefix in model_quirks.no_function_call):
        messages = [message for message in messages if message["role"] != "tool" and not "tool_calls" in message]

    # For JSON mode, the API actually requires the string "json" to be present in the user prompt
    jtag = "json:\n"
    if config.json and messages[-1]["role"] == "user" and not messages[-1]["content"].startswith(jtag):
        messages[-1]["content"] = jtag + messages[-1]["content"]

    params = {
        "model": config.model,
        "messages": messages,
    }

    if config.json and messages[-1]["role"] == "user":
        params["response_format"] = { "type": "json_object" }

    # Ugly quirk for openAI o* models: New parameter for max_tokens
    if any(config.model.startswith(prefix) for prefix in model_quirks.max_completion_tokens):
        params["max_completion_tokens"] = config.max_tokens
    else:
        params["max_tokens"] = config.max_tokens

    # Ugly quirk for openAI o* models: No temperature
    if not any(config.model.startswith(prefix) for prefix in model_quirks.no_temperature):
        params["temperature"] = config.temperature

    # Ugly quirk for openAI o1-preview* model: Does not know function calls
    if add_functions and config.func and not any(config.model.startswith(prefix) for prefix in model_quirks.no_function_call):
        params["tools"] = tools
        params["tool_choice"] = "auto"

    if config.prefix:
        params["stop"] = ["\x60\x60\x60"]

    # Try to get a completion, on rate limit retry 4 times with increasing duration
    retry = 0
    while True:
        try:
            return openai.chat.completions.create(**params)
        except openai.APIConnectionError as e:
            print(f"Failed to connect to API: {e}", file=sys.stderr)
            return None
        except openai.APIError as e:
            if isinstance(e.body, dict) and e.body["type"] == "server_error" and "message" in e.body and "tools" in e.body["message"]:
                print(f"[{config.api_provider} does not support function calling via \"tools\"]")
                config.func = False
                return create_completion_response(config, messages, False)
            else:
                print(f"API returned an API Error: {e}", file=sys.stderr)
            return None
        except openai.RateLimitError as e:
            print(f"API request exceeded rate limit: {e}", file=sys.stderr)
            retry += 1
            if retry == 5:
                return None
            else:
                time.sleep(retry ** 2)


def handle_tool_calls(config, messages, response, do_print):
    """
    Handles tool calls in the assistant's response and updates the messages list.
    Returns the final response after all tool calls are processed.
    """
    while response.choices[0].finish_reason == "tool_calls":
        assistant_message = response.choices[0].message
        tool_calls = assistant_message.tool_calls

        # Append the assistant's message with tool_calls to the messages list
        messages.append({
            "role": "assistant",
            "content": None,  # Content is None when tool_calls are present
            "tool_calls": tool_calls
        })

        # Handle each tool call
        for tool_call in tool_calls:
            handle_tool_call(messages, tool_call, do_print)

        # Make a follow-up API call with the updated messages
        allow_looped_calls = not any(config.model.startswith(prefix) for prefix in model_quirks.no_function_call_looping)
        response = create_completion_response(config, messages, allow_looped_calls)

    return response


def handle_prefix(config, messages, extend_content=False):
    if config.prefix:
        messages.append({"role": "assistant", "content": f"\x60\x60\x60{config.prefix}\n", "prefix": True})
        response = create_completion_response(config, messages, False)
        if extend_content and response:
            response.choices[0].message.content = f"\x60\x60\x60{config.prefix}\n{response.choices[0].message.content}\x60\x60\x60"
        messages[-1].pop("prefix")  # deepseek allows only the very last message to have prefix set, so delete it from message history
    else:
        response = create_completion_response(config, messages, True)

    return response


def run_ocr(config, file_paths):
    system_message = (
        "You are an automated OCR machine without user interaction. "
        "You do nothing but OCR. "
        "You output nothing but the recognized text. "
        "If the image does not contain text, reply with 'None'. "
        "Do not put the recognized text in a markdown block."
    )

    def ocr(file_content):
        if len(file_content) < 12:
            print(f"Error: File too short (length: {len(file_content)})", file=sys.stderr)
            return

        header = file_content[:12]
        if header.startswith(b'\x89PNG\r\n\x1a\n'):
            mime_image = 'image/png'
        elif header.startswith(b'\xff\xd8\xff'):
            mime_image = 'image/jpeg'
        elif header.startswith(b'RIFF') and header[8:12] == b'WEBP':
            mime_image = 'image/webp'
        elif header.startswith(b'GIF87a') or header.startswith(b'GIF89a'):
            mime_image = 'image/gif'
        else:
            print("Error: Unknown file format", file=sys.stderr)
            return

        base64_image = base64.b64encode(file_content).decode()

        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content":
                [
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{mime_image};base64,{base64_image}",
                            "detail": "auto"
                        }
                    }
                ]
            }
        ]

        response = create_completion_response(config, messages, False)

        if not (response and response.choices):
            sys.exit(1)

        output = response.choices[0].message.content
        if output != "None":
            output += '\n'
            sys.stdout.write(output)

    if not file_paths:
        ocr(sys.stdin.read())
    else:
        for file in file_paths:
            ocr(get_file_content(file, True))


def run_oneshot(config, prompt):
    system_message = (
        "You are an assistant providing a direct answer based on the user's input. "
        "You will provide a complete response immediately, without asking clarifying questions. "
        "Your reply should be clear and concise. "
        "Your reply is one-shot, i.e. not part of a conversation."
    )

    messages = [
        {"role": "system", "content": system_message},
        {"role": "user", "content": prompt}
    ]

    response = handle_prefix(config, messages, extend_content=True)

    if not (response and response.choices):
        sys.exit(1)

    response = handle_tool_calls(config, messages, response, False)

    output = response.choices[0].message.content
    print_fancy(output, config.do_markdown, config.do_latex, config.do_color, config.color)


def get_file_content(file_name, binary=False):
    if not os.path.isfile(file_name):
        print(f"Error: {file_name} does not exist.", file=sys.stderr)
        sys.exit(1)
    with open(file_name, f"r{'b' if binary else ''}") as f:
        try:
            return f.read()
        except OSError as e:
            print(f"Error opening {file_name}: {e}", file=sys.stderr)
            sys.exit(1)


def run_transform(config, prompt, file_paths):
    system_message = (
        "You are a file transformation assistant. "
        "You receive the content of a file. "
        "Execute instructions on the file content and output only the transformed, whole file. "
        "Do not add any explanations, commentary, markdown decoration such as \"\x60\x60\x60\" or other additional text. "
        f"The instructions are:\n{prompt}"
    )

    def transform(file_content):
        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": file_content}
        ]

        response = handle_prefix(config, messages)

        if not (response and response.choices):
            sys.exit(1)

        response = handle_tool_calls(config, messages, response, False)

        output = response.choices[0].message.content

        # end output with newline if the input does
        if file_content.endswith('\n'):
            output = output.rstrip('\n') + '\n'
        else:
            output = output.rstrip('\n')

        sys.stdout.write(output)

    if not file_paths:
        transform(sys.stdin.read())
    else:
        for file in file_paths:
            transform(get_file_content(file))


def run_perform(config, prompt, file_paths):
    system_message = prompt

    def perform(file_content):
        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": file_content}
        ]

        response = handle_prefix(config, messages)

        if not (response and response.choices):
            sys.exit(1)

        response = handle_tool_calls(config, messages, response, False)

        output = response.choices[0].message.content

        output += '\n'

        sys.stdout.write(output)

    if not file_paths:
        perform(sys.stdin.read())
    else:
        for file in file_paths:
            perform(get_file_content(file))


def get_multiline_input():
    print(f"[Enter multiline input (press {'Ctrl+Z' if sys.platform == 'win32' else 'Ctrl+D'} to end)]")
    lines = []
    try:
        while True:
            line = input(">>")
            lines.append(line)
    except EOFError:
        print()
    return "\n".join(lines)


def get_api_providers():
    """Extract API provider names from the config file."""
    config_file_data = load_config(config_path)
    if config_file_data and "api_providers" in config_file_data:
        return [provider["name"] for provider in config_file_data["api_providers"]]
    return []

def custom_serializer(obj):
    if isinstance(obj, ChatCompletionMessageToolCall):
        return {
            "id": obj.id,
            "type": obj.type,
            "function": {
                "name": obj.function.name,
                "arguments": obj.function.arguments
            }
        }
    # Add other custom serializations if needed
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


def run_chat(config):
    system_message = (
        "You are an assistant. "
        "You will hold a conversation with the user. "
        "Respond clearly and concisely to each user message. "
        "Avoid continuation prompts and other open ended closing statements."
    )

    messages = [{"role": "system", "content": system_message}]

    auto_completer = AutoCompleter(get_api_providers())
    auto_completer.setup()  # alters global state of readline

    print("[Entering chat:  /q to quit, /r to reset, /b for multi line buffer, /h for help]")
    while True:
        try:
            user_input = input("> ")
        except EOFError:
            print()
            continue
        except KeyboardInterrupt:
            break

        if user_input.strip() == "/h":
            print(
                "/a try prompt again\n"
                "/b multi line input buffer\n"
                "/c <file_path> save latest code block\n"
                "/d <file_path> dump chat history\n"
                "/f <file path> input file\n"
                "/h help\n"
                "/l list models\n"
                "/m <model> switch model\n"
                "/p <provider> switch API provider\n"
                "/q quit\n"
                "/r reset chat history\n"
                "/s show provider and model\n"
                "/t truncate last round\n"
                "/u <file_path> un-dump (load) chat history\n"
                "/w <float> set warmth (temperature)\n"
                "/0 reset ANSII colors ('[0m')\n"
            )
            continue

        elif user_input.strip() == "/q":
            break

        elif user_input.strip() == "/r":
            print("[Conversation history reset]\n")
            messages = [{"role": "system", "content": system_message}]
            continue

        elif user_input.strip() == "/a":
            if len(messages) > 2 and messages[-1]["role"] == "assistant" and messages[-2]["role"] == "user":
                messages.pop()  # pop reply
                user_input = messages[-1]["content"]  # get prompt as input
                messages.pop()  # pop prompt
            elif len(messages) > 1 and messages[-1]["role"] == "user":
                user_input = messages[-1]["content"]  # get prompt as input
                messages.pop()  # pop prompt
            else:
                print("[No suitable messages found]")
                continue

        elif user_input.strip().startswith("/c "):
            file_path = user_input.strip()[3:].lstrip()
            code_block = None

            for message in reversed(messages):
                if "\x60\x60\x60" in message["content"]:
                    parts = message["content"].split("\x60\x60\x60")
                    # The code block is the second-to-last segment
                    if len(parts) >= 2:
                        code_block = parts[-2].strip()
                        break

            if code_block:
                try:
                    with open(file_path, "w") as f:
                        f.write(code_block)
                    print(f"[Code block saved to {file_path}]")
                except OSError as e:
                    print(f"[Error saving code block: {e}]")
            else:
                print("[No code block found in chat history]")
            continue

        elif user_input.strip().startswith("/d "):
            file_path = user_input.strip()[3:].lstrip()
            try:
                with open(file_path, "w") as f:
                    for message in messages:
                        f.write(f"{json.dumps(message, default=custom_serializer)}\n")
                print(f"[Chat history saved to {file_path}]")
            except OSError as e:
                print(f"[Error saving chat history: {e}]")
            continue

        elif user_input.strip().startswith("/f "):
            file_path = user_input.strip()[3:].lstrip()
            try:
                with open(file_path, "r") as f:
                    file_content = f.read()
                user_input = file_content
                print(f"[File content loaded from {file_path}]")
                messages.append({"role": "user", "content": user_input})
            except OSError as e:
                print(f"[Error loading file: {e}]")
            continue

        elif user_input.strip() == "/l":
            try:
                model_ids = [model.id for model in openai.models.list()]
                print("[" + ", ".join(model_ids) + "]")
            except Exception as e:
                print(f"[Error listing models: {e}]")
            continue

        elif user_input.strip().startswith("/m "):
            new_model = user_input.strip()[3:].lstrip()
            if new_model in [model.id for model in openai.models.list()]:
                config.model = new_model
                print(f"[Model switched to {new_model}]")
            else:
                print(f"[Model {new_model} not found]")
            continue

        elif user_input.strip().startswith("/p "):
            old_config = config
            new_provider = user_input.strip()[3:].lstrip()
            config.api_provider = new_provider
            config_file_data = load_config(config_path)
            setup_config_file_base_url(config, config_file_data)
            if not config.base_url:
                config = old_config
                print(f"[no base-url found for {new_provider}]")
                continue
            setup_config_file_api_key(config, config_file_data)
            if not config.api_key:
                config = old_config
                print(f"[no api-key found for {new_provider}]")
                continue
            setup_config_file_model(config, config_file_data)
            if config.model is None:
                config = old_config
                print(f"[no model found for {new_provider}]")
                continue
            setup_config_file_func(config, config_file_data)
            openai.base_url = config.base_url.rstrip('/') + '/'
            openai.api_key = config.api_key
            # deepseek has 43 byte tool.id while openai allows max. 40
            truncate_tool_call_ids(messages)
            print(f"[API provider switched to {new_provider}]")
            continue

        elif user_input.strip() == "/s":
            print(f"{config.api_provider}/{config.model}")
            continue

        elif user_input.strip() == "/t":
            if len(messages) > 2 and messages[-1]["role"] == "assistant" and messages[-2]["role"] == "user":
                messages.pop()  # pop reply
                messages.pop()  # pop prompt
                print("[Last round truncated]")
            elif len(messages) > 1 and messages[-1]["role"] == "user":
                messages.pop()  # pop prompt
                print("[Last prompt truncated]")
            else:
                print("[No rounds to truncate]")
            continue

        elif user_input.strip().startswith("/w "):
            try:
                new_temperature = float(user_input.strip()[3:].lstrip())
                if not (0.0 <= new_temperature <= 2.0):
                    raise ValueError
                config.temperature = new_temperature
                print(f"[Temperature set to {new_temperature}]")
            except ValueError:
                print("[Invalid temperature value]")
            continue

        elif user_input.strip().startswith("/u "):
            file_path = user_input.strip()[3:].lstrip()
            try:
                with open(file_path, "r") as f:
                    # Read the file and parse each line as a JSON object
                    new_messages = []
                    for line in f:
                        try:
                            message = json.loads(line.strip())
                            new_messages.append(message)
                        except json.JSONDecodeError as e:
                            print(f"[Error decoding JSON line: {e}]")
                            continue

                    # Replace the current chat history with the loaded messages
                    if new_messages:
                        messages = new_messages
                        # deepseek has 43 byte tool.id while openai allows max. 40
                        truncate_tool_call_ids(messages)
                        print(f"[Chat history loaded from {file_path}]")
                    else:
                        print(f"[No valid chat history found in {file_path}]")
            except FileNotFoundError:
                print(f"[File not found: {file_path}]")
            except OSError as e:
                print(f"[Error loading chat history: {e}]")
            continue

        elif user_input.strip() == "/b":
            user_input = get_multiline_input()

        elif user_input.strip() == "/0":
            print_reset()
            continue

        elif user_input.strip() == "/\u0065\u0067\u0067":
            user_input = "\U0001F95A"
            print(user_input)

        messages.append({"role": "user", "content": user_input})

        response = handle_prefix(config, messages, extend_content=True)

        if not response:
            continue

        if not response.choices:
            if response.error:
                print(f"[{response.error['message']}]")
            continue

        response = handle_tool_calls(config, messages, response, True)

        output = response.choices[0].message.content
        messages.append({"role": "assistant", "content": output})
        print_fancy(output, config.do_markdown, config.do_latex, config.do_color, config.color)

    print("\n[Exiting chat]")


def setup_config_file_base_url(config, config_file_data):
    config.base_url = None

    if config_file_data and "api_providers" in config_file_data and config_file_data["api_providers"]:
        for provider in config_file_data["api_providers"]:
            if provider["name"] == config.api_provider:
                config.base_url = provider.get("base_url", None)
                break
    if not config.base_url:
        # fallback guesswork attempts
        if config.api_provider in fallbacks.base_urls:
            config.base_url = fallbacks.base_urls[config.api_provider]
            print(f"base-url not configured, trying {config.base_url} for {config.api_provider}", file=sys.stderr)


def setup_config_file_api_key(config, config_file_data):
    config.api_key = None

    if config_file_data and "api_providers" in config_file_data and config_file_data["api_providers"]:
        for provider in config_file_data["api_providers"]:
            if provider["name"] == config.api_provider:
                config.api_key = provider.get("api_key", None)
                break
    if not config.api_key:
        # fallback guesswork attempts
        if config.api_provider in fallbacks.api_keys:
            config.api_key = fallbacks.api_keys[config.api_provider]
            print(f"api_key not configured, trying {config.api_key} for {config.api_provider}", file=sys.stderr)


def setup_config_file_model(config, config_file_data):
    config.model = None

    if config_file_data and "api_providers" in config_file_data and config_file_data["api_providers"]:
        for provider in config_file_data["api_providers"]:
            if provider["name"] == config.api_provider:
                config.model = provider.get("model", None)
                break
    if config.model is None:
        # fallback guesswork attempts
        if config.api_provider in fallbacks.models:
            config.model = fallbacks.models[config.api_provider]
            print(f"model not configured, trying {config.model if config.model else "\"\""} for {config.api_provider}", file=sys.stderr)


def setup_config_file_func(config, config_file_data):
    if config_file_data and "api_providers" in config_file_data and config_file_data["api_providers"]:
        for provider in config_file_data["api_providers"]:
            if provider["name"] == config.api_provider:
                config.func = provider.get("func", config.func)
                break


def main():
    config = Config()

    help_description = f"{app_name}: Process data from a pipe or file with an OpenAI-compatible API, or chat with it."

    help_epilog = (
        "environment variables:\n"
        "    OPENAI_API_KEY=<key>\n"
        "    OPENAI_BASE_URL=<url>\n\n"
        "configuration file location:\n"
        f"    {config_path}\n\n"
        "configuration precedence (descending):\n"
        "    command line, environment, configuration file"
    )

    parser = argparse.ArgumentParser(description=help_description, epilog=help_epilog, formatter_class=argparse.RawDescriptionHelpFormatter)

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument("-c", "--chat", action="store_true", help="Interactive chat (default).")
    mode_group.add_argument("-o", "--oneshot", metavar="<prompt>", help="One-shot chat with the given prompt.")
    mode_group.add_argument("-t", "--transform", metavar="<prompt>", help="Transform data with the given prompt.")
    mode_group.add_argument("-p", "--perform", metavar="<prompt>", help="Perform task on data with the given prompt.")
    mode_group.add_argument("-l", "--list", action="store_true", help="List available models.")
    mode_group.add_argument("-r", "--ocr", action="store_true", help=argparse.SUPPRESS)
    mode_group.add_argument("-v", "--version", action="store_true", help="Print version and exit immediately.")

    parser.add_argument("-a", "--api-provider", metavar="<provider>", help="Set API provider (default: first in config file).")
    parser.add_argument("-b", "--base-url", metavar="<url>", help="Set API base URL.")
    parser.add_argument("-k", "--key", metavar="<key>", help="Set API key.")
    parser.add_argument("-m", "--model", metavar="<model>", help="Set model.")
    parser.add_argument("-x", "--max-tokens", type=int, metavar="<max>", help=f"Set maximum number of tokens (default: {config.max_tokens}).")
    parser.add_argument("-w", "--warmth", type=float, metavar="<temperature>", help=f"Set model temperature (default: {config.temperature}).")
    parser.add_argument("-j", "--json", action="store_true", help=f"Enable JSON output mode.")
    parser.add_argument("-i", "--prefix", metavar="<prefix>", help=argparse.SUPPRESS)  # deepseek beta chat prefix completion

    group_interactive = parser.add_argument_group("options for chat/oneshot mode")
    group_interactive.add_argument("--markdown", dest="markdown", action="store_true", help="Enable ANSI escape sequences for markdown.")
    group_interactive.add_argument("--no-markdown", dest="markdown", action="store_false", help="Disable ANSI escape sequences for markdown.")
    group_interactive.add_argument("--color", dest="color", action="store_true", help="Enable ANSI escape sequences for reply color.")
    group_interactive.add_argument("--no-color", dest="color", action="store_false", help="Disable ANSI escape sequences for reply color.")
    group_interactive.add_argument("--latex", dest="latex", action="store_true", help="Enable LaTeX replacement.")
    group_interactive.add_argument("--no-latex", dest="latex", action="store_false", help="Disable LaTeX replacement.")
    group_interactive.add_argument("--func", dest="func", action="store_true", help="Enable OpenAI API function calling feature.")
    group_interactive.add_argument("--no-func", dest="func", action="store_false", help="Disable OpenAI API function calling feature.")
    parser.set_defaults(markdown=None, color=None, latex=None, func=None)

    group_non_interactive = parser.add_argument_group("arguments for transform/perform mode")
    group_non_interactive.add_argument("filenames", nargs="*", help="File(s) to read from (default: stdin)")

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    env_api_key = os.getenv("OPENAI_API_KEY")
    env_base_url = os.getenv("OPENAI_BASE_URL")

    config_file_data = load_config(config_path)

    # API-provider
    if args.api_provider:
        # preset to look for in config file
        config.api_provider = args.api_provider
    elif config_file_data and "api_providers" in config_file_data and config_file_data["api_providers"]:
        # use the first preset of config file
        config.api_provider = config_file_data["api_providers"][0].get("name", None)

    # Base-URL
    if args.base_url:
        config.base_url = args.base_url
    elif env_base_url:
        config.base_url = env_base_url
    elif config.api_provider:
        setup_config_file_base_url(config, config_file_data)
    if not config.base_url:
        print("base-url not configured", file=sys.stderr)
        sys.exit(1)

    # API-key
    if args.key is not None:
        config.api_key = args.key
    elif env_api_key is not None:
        config.api_key = env_api_key
    elif config.api_provider:
        setup_config_file_api_key(config, config_file_data)
    if not config.api_key:
        print("api_key not configured", file=sys.stderr)
        sys.exit(1)

    openai.base_url = config.base_url.rstrip('/') + '/'
    openai.api_key = config.api_key

    # List
    if args.list:  # We don't need more parameters for this.
        try:
            for model in openai.models.list():
                print(model.id)
        except Exception as e:
            print(f"Error listing models: {e}", file=sys.stderr)
            sys.exit(1)
        sys.exit(0)

    # Model
    if args.model:
        config.model = args.model
    elif config.api_provider:
        setup_config_file_model(config, config_file_data)
    if config.model is None:
        print("model not configured", file=sys.stderr)
        sys.exit(1)

    # Function calling (boolean)
    if args.func is not None:
        config.func = args.func
    elif config.api_provider:
        setup_config_file_func(config, config_file_data)

    # Temperature
    if args.warmth is not None:
        config.temperature = args.warmth
    elif config_file_data and "global_options" in config_file_data:
        config.temperature = config_file_data["global_options"].get("temperature", config.temperature)

    if not 0.0 <= config.temperature <= 2.0:
        print(f"temperature {config.temperature} out of tange [0.0, 2.0]", file=sys.stderr)
        sys.exit(1)

    # Max tokens
    if args.max_tokens is not None:
        config.max_tokens = args.max_tokens
    elif config_file_data and "global_options" in config_file_data:
        config.max_tokens = config_file_data["global_options"].get("max_tokens", config.max_tokens)

    # JSON
    config.json = args.json

    # Prefix
    config.prefix = args.prefix

    # Features for chat modes
    if args.markdown is not None:
        config.do_markdown = args.markdown
    elif config_file_data and "global_options" in config_file_data:
        config.do_markdown = config_file_data["global_options"].get("markdown", config.do_markdown)

    if args.color is not None:
        config.do_color = args.color
    elif config_file_data and "global_options" in config_file_data:
        config.do_color = config_file_data["global_options"].get("color", config.do_color)

    if config_file_data and "global_options" in config_file_data:
        config.color = config_file_data["global_options"].get("color_name", config.color).lower().replace('-', '_').replace(' ', '_')

    if args.latex is not None:
        config.do_latex = args.latex
    elif config_file_data and "global_options" in config_file_data:
        config.do_latex = config_file_data["global_options"].get("latex", config.do_latex)

    del config_file_data

    # Main logic
    if args.oneshot:
        run_oneshot(config, args.oneshot)
    elif args.transform:
        run_transform(config, args.transform, args.filenames)
    elif args.perform:
        run_perform(config, args.perform, args.filenames)
    elif args.ocr:
        run_ocr(config, args.filenames)
    else:  # args.chat
        run_chat(config)


if __name__ == "__main__":
    main()
